--/**************************************************************************************************************
--*
--*   B i t H o u n d   -   A n   F P G A   B a s e d   L o g i c   A n a l y z e r
--*
--*   FPGA Design 
--* 
--* Copyright 2010-2011   Mario Mauerer (MM), Lukas Schrittwieser (LS), ETH Zurich
--*
--*    This program is free software: you can redistribute it and/or modify
--*    it under the terms of the GNU General Public License as published by
--*    the Free Software Foundation, either version 3 of the License, or
--*    (at your option) any later version.
--*
--*    This program is distributed in the hope that it will be useful,
--*    but WITHOUT ANY WARRANTY; without even the implied warranty of
--*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--*    GNU General Public License for more details.
--*
--*    You should have received a copy of the GNU General Public License
--*    along with this program.  If not, see <http://www.gnu.org/licenses/>.
--*
--***************************************************************************************************************
--*
--* Change Log:
--*
--* Version 1.0 - 2011/06/15 - LS
--*   first release
--*
--*
--***************************************************************************************************************
--*
--* Naming convention:  http://dz.ee.ethz.ch/en/information/hdl-help/vhdl-naming-conventions.html
--*
--***************************************************************************************************************
--*
--* This entity uses the IO hardware of spartan devices to synchrnoize the
--* sampling channels. It implements a simple 2-of-3 majority voting as glitch
--* filter as well as an demultiplexer to double the possible sampling rate
--*
--***************************************************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_ARITH.all;
use IEEE.STD_LOGIC_UNSIGNED.all;


library UNISIM;
use UNISIM.vcomponents.all;


entity sync is
  generic (
    CHANNELS : integer := 32);          -- number of sampling channels
  port (ClkxCI      : in  std_logic;
        ClkxCIB     : in  std_logic;    -- inverted input clock
        ChannelsxDI : in  std_logic_vector (CHANNELS-1 downto 0);  -- physical input
        -- configuration flags
        FilterxS    : in  std_logic;
        DemuxxS     : in  std_logic;
        FallingxS   : in  std_logic;
        -- synchronized data output
        SyncedxDO   : out std_logic_vector (CHANNELS-1 downto 0));
end sync;

architecture Behavioral of sync is

  -- the number after the 'S' is the pipline stage
  signal syncIn0S0xD                    : std_logic_vector(CHANNELS-1 downto 0);  -- sampled on rising edge (0 degrees)
  signal syncIn180S0xD                  : std_logic_vector(CHANNELS-1 downto 0);  -- sampled on falling edge (180
-- degrees) this value was sampled a half clock cycle earlier than syncIn0SxD
  signal syncIn0S1xDN, syncIn0S1xDP     : std_logic_vector(CHANNELS-1 downto 0);
  signal syncIn180S1xDN, syncIn180S1xDP : std_logic_vector(CHANNELS-1 downto 0);
  signal filtered0xDN, filtered0xDP     : std_logic_vector(CHANNELS-1 downto 0);
  signal filtered180xDN, filtered180xDP : std_logic_vector(CHANNELS-1 downto 0);
  signal demuxedxDN, demuxedxDP         : std_logic_vector(CHANNELS-1 downto 0);
  signal demuxedFiltxDN, demuxedFiltxDP : std_logic_vector(CHANNELS-1 downto 0);
  signal selectedxDN, selectedxDP       : std_logic_vector(CHANNELS-1 downto 0);  -- mux output
  
begin

-- use an IO DDR register to synchronize data to the rising and falling edge
-- of the sampling clock. The data synchronized to falling edge is then
-- resynchronized to the next rising edge, so evrything after the input reg is
-- in the sampling clock domain
  gen_ddr_in_reg : for i in 0 to CHANNELS-1 generate
    IDDR2_inst : IDDR2
      generic map(
        DDR_ALIGNMENT => "C0",  -- Sets output alignment to "NONE", "C0", "C1" 
        INIT_Q0       => '0',  -- Sets initial state of the Q0 output to '0' or '1'
        INIT_Q1       => '0',  -- Sets initial state of the Q1 output to '0' or '1'
        SRTYPE        => "SYNC")  -- Specifies "SYNC" or "ASYNC" set/reset
      port map (
        Q0 => syncIn0S0xD(i),           -- 1-bit output captured with C0 clock
        Q1 => syncIn180S0xD(i),         -- 1-bit output captured with C1 clock
        C0 => ClkxCI,                   -- 1-bit clock input
        C1 => ClkxCIB,                  -- 1-bit clock input
        CE => '1',                      -- 1-bit clock enable input
        D  => ChannelsxDI(i),           -- 1-bit data input 
        R  => '0',                      -- 1-bit reset input
        S  => '0'                       -- 1-bit set input
        );
  end generate gen_ddr_in_reg;

  -- TODO: Add filter for falling edge sampling (and maybe for demuxed mode too)

  -- implement majority vote (2 of 3) system as input glitch filter
  syncIn0S1xDN   <= syncIn0S0xD;        -- register stores last sample
  syncIn180S1xDN <= syncIn180S0xD;

  -- 2 of 3 majority voter as glitch filter
  gen_filter : for i in 0 to CHANNELS-1 generate
    -- the filtered signal is high if 2 of the last 3 samples (two rising, one
    -- falling edge) where high
    filtered0xDN(i)   <= (syncIn0S0xD(i) and syncIn180S0xD(i)) or (syncIn180S0xD(i) and syncIn0S1xDP(i)) or (syncIn0S0xD(i) and syncIn0S1xDP(i));
    -- the same procedure for two falling and one rising edge
    filtered180xDN(i) <= (syncIn180S0xD(i) and syncIn0S1xDP(i)) or (syncIn0S1xDP(i) and syncIn180S1xDP(i)) or (syncIn180S0xD(i) and syncIn180S1xDP(i));
  end generate gen_filter;

  -- implement a demultiplexer (16 to 32 channels) to allow  doubled sampling rate
  -- the higher 16 bits hold the older (= earlier) sample
  demuxedxDN     <= syncIn180S0xD((CHANNELS/2)-1 downto 0) & syncIn0S0xD((CHANNELS/2)-1 downto 0);
  demuxedFiltxDN <= filtered180xDP((CHANNELS/2)-1 downto 0) & filtered0xDP((CHANNELS/2)-1 downto 0);

  -- select the output signal according to the flag inputs and add an register
  process (DemuxxS, FallingxS, FilterxS, demuxedFiltxDP, demuxedxDP,
           filtered0xDP, syncIn0S1xDP, syncIn180S1xDP)
  begin
    if DemuxxS = '1' then
      if FilterxS = '1' then
        selectedxDN <= demuxedFiltxDP;
      else
        selectedxDN <= demuxedxDP;
      end if;
    elsif FilterxS = '1' then
      selectedxDN <= filtered0xDP;
    else
      if FallingxS = '1' then
        selectedxDN <= syncIn180S1xDP;
      else
        selectedxDN <= syncIn0S1xDP;
      end if;
    end if;
  end process;

  -- output the selected signal
  SyncedxDO <= selectedxDP;

  process (ClkxCI)
  begin
    if ClkxCI'event and ClkxCI = '1' then
      filtered0xDP   <= filtered0xDN;
      filtered180xDP <= filtered180xDN;
      syncIn0S1xDP   <= syncIn0S1xDN;
      syncIn180S1xDP <= syncIn180S1xDN;
      demuxedxDP     <= demuxedxDN;
      demuxedFiltxDP <= demuxedFiltxDN;
      selectedxDP    <= selectedxDN;
    end if;
  end process;
  
end Behavioral;

