--/**************************************************************************************************************
--*
--*   B i t H o u n d   -   A n   F P G A   B a s e d   L o g i c   A n a l y z e r
--*
--*   FPGA Design 
--* 
--* Copyright 2010-2011   Mario Mauerer (MM), Lukas Schrittwieser (LS), ETH Zurich
--*
--*    This program is free software: you can redistribute it and/or modify
--*    it under the terms of the GNU General Public License as published by
--*    the Free Software Foundation, either version 3 of the License, or
--*    (at your option) any later version.
--*
--*    This program is distributed in the hope that it will be useful,
--*    but WITHOUT ANY WARRANTY; without even the implied warranty of
--*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--*    GNU General Public License for more details.
--*
--*    You should have received a copy of the GNU General Public License
--*    along with this program.  If not, see <http://www.gnu.org/licenses/>.
--*
--***************************************************************************************************************
--*
--* Change Log:
--*
--* Version 1.0 - 2011/06/15 - LS
--*   first release
--*
--*
--***************************************************************************************************************
--*
--* This is a multiunit configurable trigger.it. A configurable number of
--* individual trigger units is chained to form a complex trigger.  
--*
--***************************************************************************************************************

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_ARITH.all;

entity trigger is
  generic (
    DELAY_LEN : integer := 16;  -- length of the match delay counters in bits
    UNITS_LEN : integer := 3);  -- Number of chained trigger stages (expressed as
-- power of 2, i.e. 3 equals 8 units
  port (
    ClkxCI           : in  std_logic;
    RstxRI           : in  std_logic;
    -- data input
    DatxDI           : in  std_logic_vector (31 downto 0);
    DatValidxSI      : in  std_logic;
    DemuxedxSI       : in  std_logic;
    -- trigger configuration interface
    WriteUnitxSI     : in  std_logic;   -- select the unit to be configured
    CfgDatxDI        : in  std_logic_vector (31 downto 0);
    WriteMaskxSI     : in  std_logic;
    WriteMatchValxSI : in  std_logic;
    WriteFlagsxSI    : in  std_logic;
    WriteDelayxSI    : in  std_logic;
    ArmxSI           : in  std_logic;   -- activates the trigger
    RunxSO           : out std_logic;   -- goes to high when the trigger fires
    -- dataoutput (pipelining!)
    DatxDO           : out std_logic_vector(31 downto 0);
    DatValidxSO      : out std_logic
    );
end trigger;

architecture Behavioral of trigger is

  component triggerUnit
    generic (
      DELAY_LEN : integer);
    port (
      RstxR         : in  std_logic;
      ClkxCI        : in  std_logic;
      MaskxDI       : in  std_logic_vector(31 downto 0);
      MatchValxDI   : in  std_logic_vector(31 downto 0);
      DemuxedxSI    : in  std_logic;
      CfgChannelxDI : in  std_logic_vector(4 downto 0);
      CfgSerialxSI  : in  std_logic;
      CfgNotxSI     : in  std_logic;
      CfgChainxSI   : in  std_logic;
      CfgNextLvlxSI : in  std_logic;
      CfgFirexSI    : in  std_logic;
      CfgDelayxDI   : in  std_logic_vector(DELAY_LEN-1 downto 0);
      DInxDI        : in  std_logic_vector(31 downto 0);
      DInValidxSI   : in  std_logic;
      MatchChainxSI : in  std_logic;
      FiredxSI      : in  std_logic;
      ActivexSI     : in  std_logic;
      DOutxDO       : out std_logic_vector(31 downto 0);
      DOutValidxSO  : out std_logic;
      MatchChainxSO : out std_logic;
      FiredxSO      : out std_logic;
      ActivexSO     : out std_logic);
  end component;

  constant NUM_UNITS : integer := 2**UNITS_LEN;

  -- flag bit positions
  constant FLAG_POS_SERIAL   : integer := 8;
  constant FLAG_POS_NOT      : integer := 9;
  constant FLAG_POS_CHAIN    : integer := 10;
  constant FLAG_POS_NEXT_LVL : integer := 11;
  constant FLAG_POS_FIRE     : integer := 12;


  type CfgRegType is array (0 to NUM_UNITS-1) of std_logic_vector(31 downto 0);  -- type for configuration registers
  type DataChainType is array (0 to NUM_UNITS-1) of std_logic_vector(31 downto 0);  -- type for chaining the units

  signal SelUnitxDN, SelUnitxDP   : std_logic_vector(UNITS_LEN-1 downto 0) := (others => '0');
  signal MaskxDN, MaskxDP         : CfgRegType;
  signal MatchValxDN, MatchValxDP : CfgRegType;
  signal DelayxDN, DelayxDP       : CfgRegType;
  signal FlagsxDN, FlagsxDP       : CfgRegType;

  signal TriggerActivexSN, TriggerActivexSP : std_logic := '0';

  -- two signals are used to chain the trigger units, this simplifies the
-- generate statements
  signal DataInxD       : DataChainType;
  signal DataOutxD      : DataChainType;
  signal DataInValidxS  : std_logic_vector(0 to NUM_UNITS-1);
  signal DataOutValidxS : std_logic_vector(0 to NUM_UNITS-1);
  signal MatchInxS      : std_logic_vector(0 to NUM_UNITS-1);
  signal MatchOutxS     : std_logic_vector(0 to NUM_UNITS-1);
  signal FiredInxS      : std_logic_vector(0 to NUM_UNITS-1);
  signal FiredOutxS     : std_logic_vector(0 to NUM_UNITS-1);
  signal ActiveInxS     : std_logic_vector(0 to NUM_UNITS-1);
  signal ActiveOutxS    : std_logic_vector(0 to NUM_UNITS-1);
  
begin

  -- create stages
  TUnits : for i in 0 to NUM_UNITS-1 generate
    triggerUnitInst : triggerUnit
      generic map (
        DELAY_LEN => DELAY_LEN)
      port map (
        RstxR         => RstxRI,
        ClkxCI        => ClkxCI,
        MaskxDI       => MaskxDP(i),
        MatchValxDI   => MatchValxDP(i),
        DemuxedxSI    => DemuxedxSI,
        CfgChannelxDI => FlagsxDP(i)(4 downto 0),
        CfgSerialxSI  => FlagsxDP(i)(FLAG_POS_SERIAL),
        CfgNotxSI     => FlagsxDP(i)(FLAG_POS_NOT),
        CfgChainxSI   => FlagsxDP(i)(FLAG_POS_CHAIN),
        CfgNextLvlxSI => FlagsxDP(i)(FLAG_POS_NEXT_LVL),
        CfgFirexSI    => FlagsxDP(i)(FLAG_POS_FIRE),
        CfgDelayxDI   => DelayxDP(i)(DELAY_LEN-1 downto 0),
        DInxDI        => DataInxD(i),
        DInValidxSI   => DataInValidxS(i),
        MatchChainxSI => MatchInxS(i),
        FiredxSI      => FiredInxS(i),
        ActivexSI     => ActiveInxS(i),
        DOutxDO       => DataOutxD(i),
        DOutValidxSO  => DataOutValidxS(i),
        MatchChainxSO => MatchOutxS(i),
        FiredxSO      => FiredOutxS(i),
        ActivexSO     => ActiveOutxS(i));

  end generate;

  -- chain units by connecting input and output signals
  chain : for i in 1 to NUM_UNITS-1 generate
    DataInxD(i)      <= DataOutxD(i-1);
    DataInValidxS(i) <= DataOutValidxS(i-1);
    MatchInxS(i)     <= MatchOutxS(i-1);
    FiredInxS(i)     <= FiredOutxS(i-1);
    ActiveInxS(i)    <= ActiveOutxS(i-1);
  end generate chain;


  -- input assignemnts for first unit
  DataInxD(0)      <= DatxDI;
  DataInValidxS(0) <= DatValidxSI;
  MatchInxS(0)     <= '0';
  FiredInxS(0)     <= '0';
  -- trigger starts with Arm signal and stays active until resetted
  ActiveInxS(0)    <= TriggerActivexSP or ArmxSI;
  TriggerActivexSN <= TriggerActivexSP or ArmxSI;

  -- output assignments for last unit
  DatxDO      <= DataOutxD(NUM_UNITS-1);
  DatValidxSO <= DataOutValidxS(NUM_UNITS-1);
  RunxSO      <= FiredOutxS(NUM_UNITS-1);
  
  
  process (CfgDatxDI, DelayxDP, FlagsxDP, MaskxDP, MatchValxDP, SelUnitxDP,
           WriteDelayxSI, WriteFlagsxSI, WriteMaskxSI, WriteMatchValxSI,
           WriteUnitxSI)
    variable selUnit : integer;
  begin
    -- default: do nothing
    SelUnitxDN  <= SelUnitxDP;
    MaskxDN     <= MaskxDP;
    MatchValxDN <= MatchValxDP;
    FlagsxDN    <= FlagsxDP;
    DelayxDN    <= DelayxDP;
    if WriteUnitxSI = '1' then
      SelUnitxDN <= CfgDatxDI(UNITS_LEN-1 downto 0);
    end if;
    -- assign selected unit to a variable for better readability
    selUnit := conv_integer(unsigned(SelUnitxDP));
    if WriteMaskxSI = '1' then
      MaskxDN(selUnit) <= CfgDatxDI;
    end if;
    if WriteMatchValxSI = '1' then
      MatchValxDN(selUnit) <= CfgDatxDI;
    end if;
    if WriteDelayxSI = '1' then
      DelayxDN(selUnit) <= CfgDatxDI;
    end if;
    if WriteFlagsxSI = '1' then
      FlagsxDN(selUnit)                    <= CfgDatxDI;
      -- overwrite flags for the first and last unit as they don't make sense
      FlagsxDN(0)(FLAG_POS_CHAIN)          <= '0';  -- can't chain first unit
      FlagsxDN(0)(FLAG_POS_NEXT_LVL)       <= '0';  -- unit is started with active
      --FlagsxDN(NUM_UNITS-1)(FLAG_POS_FIRE) <= '1';  -- last unit always fires
    end if;
  end process;


  regs : process (ClkxCI)
  begin  -- process regs
    if ClkxCI'event and ClkxCI = '1' then  -- rising clock edge
      if RstxRI = '1' then
        SelUnitxDP       <= (others => '0');
        TriggerActivexSP <= '0';
        MaskxDP          <= (others => (others => '0'));
        MatchValxDP      <= (others => (others => '0'));
        DelayxDP         <= (others => (others => '0'));
        FlagsxDP         <= (others => (others => '0'));
      else
        SelUnitxDP       <= SelUnitxDN;
        TriggerActivexSP <= TriggerActivexSN;
        MaskxDP          <= MaskxDN;
        MatchValxDP      <= MatchValxDN;
        DelayxDP         <= DelayxDN;
        FlagsxDP         <= FlagsxDN;
      end if;
    end if;
  end process regs;
  

end Behavioral;

