--/**************************************************************************************************************
--*
--*   B i t H o u n d   -   A n   F P G A   B a s e d   L o g i c   A n a l y z e r
--*
--*   FPGA Design 
--* 
--* Copyright 2010-2011   Mario Mauerer (MM), Lukas Schrittwieser (LS), ETH Zurich
--*
--*    This program is free software: you can redistribute it and/or modify
--*    it under the terms of the GNU General Public License as published by
--*    the Free Software Foundation, either version 3 of the License, or
--*    (at your option) any later version.
--*
--*    This program is distributed in the hope that it will be useful,
--*    but WITHOUT ANY WARRANTY; without even the implied warranty of
--*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--*    GNU General Public License for more details.
--*
--*    You should have received a copy of the GNU General Public License
--*    along with this program.  If not, see <http://www.gnu.org/licenses/>.
--*
--***************************************************************************************************************
--*
--* Change Log:
--*
--* Version 1.0 - 2011/06/15 - LS
--*   first release
--*
--*
--***************************************************************************************************************
--*
--* This is a chainable, configurable pipelined trigger unit. Mulitplethe
--* instances can be chained to form a complex trigger. See the BitHound
--* documentation for a detailed explanation.  
--*
--***************************************************************************************************************


library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.std_logic_arith.all;

entity triggerUnit is
  generic (
    DELAY_LEN : integer := 16
    );
  port (RstxR         : in std_logic;
        ClkxCI        : in std_logic;
        -- configuration interface
        MaskxDI       : in std_logic_vector(31 downto 0);
        MatchValxDI   : in std_logic_vector(31 downto 0);
        DemuxedxSI    : in std_logic;
        CfgChannelxDI : in std_logic_vector(4 downto 0);
        CfgSerialxSI  : in std_logic;
        CfgNotxSI     : in std_logic;
        CfgChainxSI   : in std_logic;
        CfgNextLvlxSI : in std_logic;
        CfgFirexSI    : in std_logic;
        CfgDelayxDI   : in std_logic_vector(DELAY_LEN-1 downto 0);

        -- trigger pipeline inputs
        DInxDI        : in std_logic_vector(31 downto 0);
        DInValidxSI   : in std_logic;
        MatchChainxSI : in std_logic;
        FiredxSI      : in std_logic;
        ActivexSI     : in std_logic;

        -- trigger pipeline outputs
        DOutxDO       : out std_logic_vector(31 downto 0);
        DOutValidxSO  : out std_logic;
        MatchChainxSO : out std_logic;
        FiredxSO      : out std_logic;
        ActivexSO     : out std_logic

        );
end triggerUnit;

architecture Behavioral of triggerUnit is

  -- pipeline stage 0
  signal InRegxDN, InRegxDP                   : std_logic_vector(31 downto 0);
  signal InShiftxDN, InShiftxDP               : std_logic_vector(31 downto 0);
  signal InDemuxedShiftxDN, InDemuxedShiftxDP : std_logic_vector(32 downto 0);
--      signal InDemuxedShift1xDN, InDemuxedShift1xDP : std_logic_vector(31 downto 0);
  signal DValid0xSN, DValid0xSP               : std_logic                     := '0';
  signal Data0xDN, Data0xDP                   : std_logic_vector(31 downto 0) := x"00000000";
  signal MatchChain0xSN, MatchChain0xSP       : std_logic                     := '0';
  signal Fired0xSN, Fired0xSP                 : std_logic                     := '0';
  signal Active0xSN, Active0xSP               : std_logic                     := '0';

  -- pipeline stage 1
  signal MatchParxSN, MatchParxSP           : std_logic;
  signal MatchSerxSN, MatchSerxSP           : std_logic;
  signal MatchDemuxParxSN, MatchDemuxParxSP : std_logic;
  signal MatchDemuxSerxSN, MatchDemuxSerxSP : std_logic;
  signal DValid1xSN, DValid1xSP             : std_logic                     := '0';
  signal Data1xDN, Data1xDP                 : std_logic_vector(31 downto 0) := x"00000000";
  signal MatchChain1xSN, MatchChain1xSP     : std_logic                     := '0';
  signal Fired1xSN, Fired1xSP               : std_logic                     := '0';
  signal Active1xSN, Active1xSP             : std_logic                     := '0';

  -- pipeline stage 2
  signal MatchxS                        : std_logic;
  signal DValid2xSN, DValid2xSP         : std_logic                     := '0';
  signal Data2xDN, Data2xDP             : std_logic_vector(31 downto 0) := x"00000000";
  signal MatchChain2xSN, MatchChain2xSP : std_logic                     := '0';
  signal Fired2xSN, Fired2xSP           : std_logic                     := '0';
  signal Active2xSN, Active2xSP         : std_logic                     := '0';

  -- pipeline stage 3
  signal DelayCntxDN, DelayCntxDP         : unsigned(DELAY_LEN-1 downto 0);
  signal DelayCntRunxSN, DelayCntRunxSP   : std_logic                     := '0';
  signal DelayedMatchxSN, DelayedMatchxSP : std_logic                     := '0';
  signal DValid3xSN, DValid3xSP           : std_logic                     := '0';
  signal Data3xDN, Data3xDP               : std_logic_vector(31 downto 0) := x"00000000";
  signal MatchChain3xSN, MatchChain3xSP   : std_logic                     := '0';
  signal Fired3xSN, Fired3xSP             : std_logic                     := '0';
  signal Active3xSN, Active3xSP           : std_logic                     := '0';


  -- pipeline stage 4
  signal DValid4xSN, DValid4xSP         : std_logic                     := '0';
  signal Data4xDN, Data4xDP             : std_logic_vector(31 downto 0) := x"00000000";
  signal MatchChain4xSN, MatchChain4xSP : std_logic                     := '0';
  signal Fired4xSN, Fired4xSP           : std_logic                     := '0';
  signal Active4xSN, Active4xSP         : std_logic                     := '0';
  
begin

-- pipeline stage 0: parallel input register, shift registers
  DValid0xSN     <= DinValidxSI;
  Data0xDN       <= DInxDI;
  MatchChain0xSN <= MatchChainxSI;
-- split the fired signal if this is the beginning of a new level
  Fired0xSN      <= '0' when CfgNextLvlxSI = '1' else FiredxSI;
-- select activation signal of the this stage according to its configured level
  Active0xSN     <= FiredxSI when CfgNextLvlxSI = '1' else ActivexSI;

  process(CfgChannelxDI, DInxDI, DinValidxSI, InDemuxedShiftxDP, InRegxDP,
          InShiftxDP)
    variable selChxD        : integer range 0 to 31;
    variable selChDemuxedxD : integer range 0 to 15;
  begin
    -- assign selected channel to variable for better readability
    selChxD           := conv_integer(unsigned(CfgChannelxDI));
    selChDemuxedxD    := conv_integer(unsigned(CfgChannelxDI(3 downto 0)));
    -- default: do nothing
    InRegxDN          <= InRegxDP;
    InShiftxDN        <= InShiftxDP;
    InDemuxedShiftxDN <= InDemuxedShiftxDP;
    if DinValidxSI = '1' then
      InRegxDN          <= DInxDI;
      InShiftxDN        <= InShiftxDP(30 downto 0) & DInxDI(selChxD);
      -- in demuxed mode two samples are processed at the same time, so two bits are shifted in.
      InDemuxedShiftxDN <= InDemuxedShiftxDP(30 downto 0) & DInxDI(selChDemuxedxD+16) & DInxDI(selChDemuxedxD);
    end if;
  end process;

  process(ClkxCI)
  begin
    if ClkxCI'event and ClkxCI = '1' then
      
      if RstxR = '1' then
        InRegxDP          <= (others => '0');
        InShiftxDP        <= (others => '0');
        InDemuxedShiftxDP <= (others => '0');
        Data0xDP          <= (others => '0');
        DValid0xSP        <= '0';
        MatchChain0xSP    <= '0';
        Fired0xSP         <= '0';
        Active0xSP        <= '0';
      else
        InRegxDP          <= InRegxDN;
        InShiftxDP        <= InShiftxDN;
        InDemuxedShiftxDP <= InDemuxedShiftxDN;
        Data0xDP          <= Data0xDN;
        DValid0xSP        <= DValid0xSN;
        MatchChain0xSP    <= MatchChain0xSN;
        Fired0xSP         <= Fired0xSN;
        Active0xSP        <= Active0xSN;
      end if;
    end if;
  end process;


-- pipeline stage 1: mask values and compare
  DValid1xSN     <= DValid0xSP;
  Data1xDN       <= Data0xDP;
  MatchChain1xSN <= MatchChain0xSP;
  Fired1xSN      <= Fired0xSP;
  Active1xSN     <= Active0xSP;

  process(InDemuxedShiftxDP, InRegxDP, InShiftxDP, MaskxDI, MatchValxDI)
    variable MaskedParxD         : std_logic_vector(31 downto 0);
    variable MaskedSerxD         : std_logic_vector(31 downto 0);
    variable MaskedDemuxedParxD  : std_logic_vector(31 downto 0);
    variable MaskedDemuxedSer0xD : std_logic_vector(31 downto 0);
    variable MaskedDemuxedSer1xD : std_logic_vector(31 downto 0);
  begin
    -- xor the samples with the value we should check against and mask out the bits we must noch check
    -- we assign the different results to variables for better readability
    MaskedParxD         := (InRegxDP xor MatchValxDI) and MaskxDI;
    MaskedSerxD         := (InShiftxDP xor MatchValxDI) and MaskxDI;
    -- in demultiplexed mode two samples (a 16 bits) are processed at the same time
    MaskedDemuxedParxD  := (InRegxDP xor (MatchValxDI(15 downto 0)&MatchValxDI(15 downto 0))) and (MaskxDI(15 downto 0) & MaskxDI(15 downto 0));
    -- in serial demuxed mode we have to check two shift reg values. with one and with two bits shifted in
    MaskedDemuxedSer0xD := (InDemuxedShiftxDP(31 downto 0) xor MatchValxDI) and MaskxDI;
    MaskedDemuxedSer1xD := (InDemuxedShiftxDP(32 downto 1) xor MatchValxDI) and MaskxDI;
    -- now do the zero checking
    MatchParxSN         <= '0';
    if MaskedParxD = x"00000000" then
      MatchParxSN <= '1';
    end if;
    MatchSerxSN <= '0';
    if MaskedSerxD = x"00000000" then
      MatchSerxSN <= '1';
    end if;
    MatchDemuxParxSN <= '0';
    if (MaskedDemuxedParxD(15 downto 0) = x"0000") or (MaskedDemuxedParxD(31 downto 16) = x"0000") then
      MatchDemuxParxSN <= '1';
    end if;
    MatchDemuxSerxSN <= '0';
    if MaskedDemuxedSer0xD = x"00000000" or MaskedDemuxedSer1xD = x"00000000" then
      MatchDemuxSerxSN <= '1';
    end if;
  end process;

  process (ClkxCI)
  begin
    if ClkxCI'event and ClkxCI = '1' then
      Data1xDP <= Data1xDN;
      if RstxR = '1' then
        DValid1xSP       <= '0';
        MatchChain1xSP   <= '0';
        Fired1xSP        <= '0';
        Active1xSP       <= '0';
        MatchParxSP      <= '0';
        MatchSerxSP      <= '0';
        MatchDemuxSerxSP <= '0';
        MatchDemuxParxSP <= '0';
      else
        DValid1xSP       <= DValid1xSN;
        MatchChain1xSP   <= MatchChain1xSN;
        Fired1xSP        <= Fired1xSN;
        Active1xSP       <= Active1xSN;
        MatchParxSP      <= MatchParxSN;
        MatchSerxSP      <= MatchSerxSN;
        MatchDemuxSerxSP <= MatchDemuxSerxSN;
        MatchDemuxParxSP <= MatchDemuxParxSN;
      end if;
    end if;
  end process;

-- pipeline stage 2: Match signal processing
  DValid2xSN <= DValid1xSP;
  Data2xDN   <= Data1xDP;
  -- MatchChain2xSN <= MatchChain1xSP;
  Fired2xSN  <= Fired1xSP;
  Active2xSN <= Active1xSP;

-- select match signal according to configuration       
  process(CfgSerialxSI, DValid1xSP, DemuxedxSI, MatchDemuxParxSP,
          MatchDemuxSerxSP, MatchParxSP, MatchSerxSP)
  begin
    MatchxS <= '0';
    if DValid1xSP = '1' then
      if DemuxedxSI = '0' then
        if CfgSerialxSI = '1' then
          MatchxS <= MatchSerxSP;
        else
          MatchxS <= MatchParxSP;
        end if;
      else
        if CfgSerialxSI = '1' then
          MatchxS <= MatchDemuxSerxSP;
        else
          MatchxS <= MatchDemuxParxSP;
        end if;
      end if;
    end if;
  end process;

-- match signal chaining logic  
  process(Active1xSP, CfgChainxSI, CfgNotxSI, MatchChain1xSP, MatchxS)
  begin
    MatchChain2xSN <= '0';
    if Active1xSP = '1' then
      if CfgChainxSI = '1' then
        MatchChain2xSN <= MatchChain1xSP and (MatchxS xor CfgNotxSI);
      else
        MatchChain2xSN <= MatchxS xor CfgNotxSI;
      end if;
    end if;
  end process;


  process (ClkxCI)
  begin
    if ClkxCI'event and ClkxCI = '1' then
      Data2xDP <= Data2xDN;
      if RstxR = '1' then
        DValid2xSP     <= '0';
        MatchChain2xSP <= '0';
        Fired2xSP      <= '0';
        Active2xSP     <= '0';
      else
        DValid2xSP     <= DValid2xSN;
        MatchChain2xSP <= MatchChain2xSN;
        Fired2xSP      <= Fired2xSN;
        Active2xSP     <= Active2xSN;
      end if;
    end if;
  end process;

-- pipeline stage 3: match signal delay
  DValid3xSN     <= DValid2xSP;
  Data3xDN       <= Data2xDP;
  MatchChain3xSN <= MatchChain2xSP;
  Fired3xSN      <= Fired2xSP;
  Active3xSN     <= Active2xSP;

  process(Active2xSP, CfgDelayxDI, DValid2xSP, DelayCntRunxSP, DelayCntxDP,
          DelayedMatchxSP, MatchChain2xSP)
  begin
    -- default assignments
    DelayCntxDN     <= DelayCntxDP;
    DelayedMatchxSN <= '0';
    -- the match chain impuls starts the counter and sets the run counter signal
    DelayCntRunxSN  <= (DelayCntRunxSP or MatchChain2xSP) and Active2xSP and (not DelayedMatchxSP);
    if (DelayCntRunxSP = '1' or MatchChain2xSP = '1') and Active2xSP = '1' then
      if DValid2xSP = '1' then
        DelayCntxDN <= DelayCntxDP + 1;
      end if;
    end if;
    -- forward the match signal once the counter has reached it's final value
    if DelayCntxDP = unsigned(CfgDelayxDI) and DValid2xSP = '1' and Active2xSP = '1' then
      -- stop the counter
      DelayCntRunxSN  <= '0';
      -- MatchChain2xSP appears here to make the counter transparent if CfgDelayxDI is set to 0
      DelayedMatchxSN <= DelayCntRunxSP or MatchChain2xSP;
    end if;
  end process;

  process (ClkxCI)
  begin
    if ClkxCI'event and ClkxCI = '1' then
      if RstxR = '1' then
        Data3xDP        <= (others => '0');
        DValid3xSP      <= '0';
        MatchChain3xSP  <= '0';
        Fired3xSP       <= '0';
        Active3xSP      <= '0';
        DelayCntxDP     <= (others => '0');
        DelayCntRunxSP  <= '0';
        DelayedMatchxSP <= '0';
      else
        Data3xDP        <= Data3xDN;
        DValid3xSP      <= DValid3xSN;
        MatchChain3xSP  <= MatchChain3xSN;
        Fired3xSP       <= Fired3xSN;
        Active3xSP      <= Active3xSN;
        DelayCntxDP     <= DelayCntxDN;
        DelayCntRunxSP  <= DelayCntRunxSN;
        DelayedMatchxSP <= DelayedMatchxSN;
      end if;
    end if;
  end process;


-- pipeline stage 4: fire signal logic
  DValid4xSN     <= DValid3xSP;
  Data4xDN       <= Data3xDP;
  MatchChain4xSN <= MatchChain3xSP;
  Active4xSN     <= Active3xSP;

  Fired4xSN <= Fired4xSP or Fired3xSP or (DelayedMatchxSP and CfgFirexSI);

  process (ClkxCI)
  begin
    if ClkxCI'event and ClkxCI = '1' then
      
      if RstxR = '1' then
        Data4xDP       <= (others => '0');
        DValid4xSP     <= '0';
        MatchChain4xSP <= '0';
        Fired4xSP      <= '0';
        Active4xSP     <= '0';
      else
        Data4xDP       <= Data4xDN;
        DValid4xSP     <= DValid4xSN;
        MatchChain4xSP <= MatchChain4xSN;
        Fired4xSP      <= Fired4xSN;
        Active4xSP     <= Active4xSN;
      end if;
    end if;
  end process;

-- output assignments
  DOutxDO       <= Data4xDP;
  DOutValidxSO  <= DValid4xSP;
  MatchChainxSO <= MatchChain4xSP;
  FiredxSO      <= Fired4xSP;
  ActivexSO     <= Active4xSP;
  
  
end Behavioral;

-- end of file
