/**************************************************************************************************************
--*
--*   B i t H o u n d   -   A n   F P G A   B a s e d   L o g i c   A n a l y z e r
--*
--*   Firmware 
--* 
--* Copyright 2010-2011   Mario Mauerer (MM), Lukas Schrittwieser (LS), ETH Zurich
--*
--*    This program is free software: you can redistribute it and/or modify
--*    it under the terms of the GNU General Public License as published by
--*    the Free Software Foundation, either version 3 of the License, or
--*    (at your option) any later version.
--*
--*    This program is distributed in the hope that it will be useful,
--*    but WITHOUT ANY WARRANTY; without even the implied warranty of
--*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--*    GNU General Public License for more details.
--*
--*    You should have received a copy of the GNU General Public License
--*    along with this program.  If not, see <http://www.gnu.org/licenses/>.
--*
--***************************************************************************************************************
--*
--* Change Log:
--*
--* Version 1.0 - 2011/08/17 - LS
--*   first release
--*
--*
--***************************************************************************************************************
--*
--* This is the high level application. It processes segments passed by the TCP stack and implements the 
--* communication protocol with the client (GUI application).
--*
--***************************************************************************************************************/

#include <string.h>
#include <avr/interrupt.h>
#include "app.h"
#include "uip.h"
#include "uart.h"
#include "timer.h"
#include "dhcpc.h"
#include "io.h"

#define BROADCAST_PORT		1024
#define	TCP_PORT			1024
#define DEFAULT_DATA_PORT	1025

#define SAMPLES_PER_UDP_P	360

#define OPCODE_RESET		0x00
#define OPCODE_ARM			0x01
#define	OPCODE_ID			0x02
#define OPCODE_SET_READ_DEL	0x81
#define OPCODE_SIZE_FWD		0x83
#define OPCODE_SIZE_BWD		0x84
#define OPCODE_DATA_PORT	0x85
#define OPCODE_RESEND		0x86

#define ID_CODE_READY		"SLA2"
#define ID_CODE_ERR0    	"ERR0"

#define ST_IDLE				0
#define ST_RCV_DATA 		1

#ifndef HTONL
	#if UIP_BYTE_ORDER == UIP_BIG_ENDIAN
		#define HTONL(x) 	(x)
	#else
		#define HTONL(x) 	((((uint32_t)(HTONS((x)&0xffff)))<<16) | HTONS(((x)>>16)&0xffff)) 
	#endif 
#endif


/**************************************************************************************************
*   H A R D W A R E   P R O P E R T I E S                                          
*/

// we have 128MBytes of sample ram and one sample consists of 4 bytes (32 channels)
#define SAMPLE_MEM_LENGTH		((uint32_t)(32L*1024L*1024L))
#define SAMPLE_MEM_ADDR_MASK	(0x7ffffffL)	


#define NUM_SAMPLING_CHANNELS	32

// the base frequency of the sampling hardware in MHz
#define MAX_SAMPLING_FRQ		200

// dram dma interface
// NOTE: due to the wb-wb-bridge inthe hardware all accesses to these register have to be atomic
#define DRAM_BASE				((uint16_t)0xa000)
#define DRAM_CFG_STAT			(*((volatile uint32_t*)(DRAM_BASE+0)))
#define DRAM_SRC_ADDR			(*((volatile uint32_t*)(DRAM_BASE+4)))
#define DRAM_DEST_ADDR			(*((volatile uint32_t*)(DRAM_BASE+8)))
#define DRAM_TX_LEN				(*((volatile uint32_t*)(DRAM_BASE+12)))
#define DRAM_SAMPLING_PTR		(*((volatile uint32_t*)(DRAM_BASE+16)))
#define DRAM_BUSY				0x01
#define DRAM_IE					0x02
#define DRAM_DONE				0x04
#define DRAM_ERR				0x08



/**************************************************************************************************
*   G L O B A L S                                            
*/

/* globals for communication with uIP */
struct uip_udp_conn* 	broadcastConn = 0;
struct uip_udp_conn*	udpDataConn = 0;
struct timer 			udpTimer;
uint8_t					configured 		= 0;
struct uip_conn*		connectedClient = 0;

/* globals for communication with the sampling core and client */	
int32_t				preTrigLen = 0;		// how many bytes before the trigger should be sent to the client
int32_t				postTrigLen = 0;	// how many bytes after the trigger should be sent to the client
uint16_t			dataDestPort;		// dest. UDP port number for sample data
volatile uint8_t	hwDataReady = 0;	// 1: the hardware has finished sampling and has data to send 
volatile uint8_t	waitForTrigger  = 0;
uint32_t			samplesSent = 0;	// how many samples have been sent to the client
uint32_t			samplesToSend = 0;	// total number of samples to send to the client		

/* structure of the UDP broadcast advertisement package 
   All uint16_t and uint32_t are in network byte order */
struct advertisementPkg
{
	char		name[6];			// fixed magic string: "SumpLA"
	char		idCode[4];			// id code of the logic analyzer
	uint16_t	port;				// port number of the tcp server (where the LA awaits connections)
	uint32_t	deviceIpAddr;		// ip v4 address of the LA
	uint32_t	sampleMemLength;	// how many samples the LA can store at maximum (when using all channels)
	uint16_t	numChannels;		// how many sampling channels does the LA have?
	uint16_t	maxSamplingFrq;		// maximum sampling frequency of the LA in MHz
};



/**************************************************************************************************
*   P R O T O T Y P E S                                             
*/
void execute (struct sump_state* s);
void process(struct sump_state* s);
void handle_connection(struct sump_state *s);
uint8_t send (struct sump_state* s, const void* data, uint8_t len);



/**************************************************************************************************
*   I M P L E M E N T A T I O N 
*/

/* this gets called by the stack in any tcp related event occured */
void appcall(void)
{
	/* before the configuration all tcp frames are ignored. (the app is configured after
	   we got an ip address via dhcp) */
	if (!configured)
		return;

	struct sump_state *s = (struct sump_state *)&(uip_conn->appstate);
 
   	if (uip_closed() || uip_aborted() || uip_timedout()) 
	{
		// if the connection of the active client was closed we are free for a new client
		if (connectedClient == uip_conn)
			connectedClient=NULL;
		uartPrintStr_P(PSTR("closed\n"));
	}
	else if (uip_connected()) 
	{
		// a new client wishes to connect, prepare the data structures
		s->timer = 0;
		s->txBufWr = 0;
		s->txBufRd = 0;
		s->txSegLen = 0;
		s->dataRcvCnt = 0;
		s->state = ST_IDLE;
		uartPrintStr_P(PSTR("connected\n"));
		// The first client automatically gets access to the hardware
		if (connectedClient==NULL)
			connectedClient = uip_conn;		
     	handle_connection(s);
   	}
	else if(s != 0)
	{
		// if a connection is idle for too long we close it. This timeout
		// prevents that a crashed client which has not closed the connection
		// blocks the firmware
    	if(uip_poll() && (!waitForTrigger))
		{
			uartPutc('.');
       		s->timer++;
       		if(s->timer >= 60) 
			{
				uartPrintStr_P(PSTR("60s idle: abort\n"));
         		uip_close();
       		}
     	}
		else
		{
       		s->timer = 0;
     	}
		// this is a call for special connection, do the data reception and transmission
     	handle_connection(s);
   	}
   	else
   	{
		uartPrintStr_P(PSTR("s=0: abort\n"));
   		uip_abort();
   	}

}


void handle_connection(struct sump_state *s)
{
	uint8_t i;

	// check if we should retransmit the last packet
	if (uip_rexmit())
	{
		uartPrintStr_P(PSTR("retrans\n"));
		if (s->txSegLen == 0)
		{
			// the stack requests a retransmission but we have no segment to retransmit, this is an error
			uartPrintStr_P(PSTR("err: no retrans data, closing conn\n"));
			uip_close();
		}
		uip_send(s->txSeg, s->txSegLen);
	}
	if (uip_acked())
	{
		uartPrintStr_P(PSTR("got ack\n"));
		s->txSegLen = 0;
	}
	// check if we got new data
	if (uip_newdata())
	{
		// process the received data if
		process(s);
	}
	// do we have new data to send and is there no segment in transmission?
	if ((s->txBufRd != s->txBufWr) && (s->txSegLen==0))
	{
		uartPrintStr("send buf\n");
		// we copy the data from the ring-buffer into the segment buffer
		for (i=0; (i<TCP_TX_LEN) && (s->txBufRd != s->txBufWr); i++)
		{
			s->txSeg[i] = s->txBuf[(s->txBufRd)++]; // copy the byte
			// check read pointer overrun
			if ((s->txBufRd) >= BUFFER_LENGTH)
				s->txBufRd = 0;
		}
		s->txSegLen = i;
		uip_send(s->txSeg, s->txSegLen);
		//s->waitForAck=1;
	}
}


void process(struct sump_state* s)
{
	uint8_t i;
	uint8_t d; 	// data byte

	for (i=0; i<uip_datalen();i++)
	{
        d = ((uint8_t*)uip_appdata)[i];		

		switch (s->state)
		{
			case ST_IDLE:	// first byte is the opcode
				s->opcode = d;
				if ((s->opcode) & 0x80)
				{
					// this is a command with data -> we need four more bytes
					s->state = ST_RCV_DATA;
					s->dataRcvCnt = 0;
				}
				else
				{
					// this is a command with only an opcode -> execute it
					execute(s);
				}
				break;
			
			case ST_RCV_DATA:
				s->data[s->dataRcvCnt++] = d;
				if (s->dataRcvCnt == 4)
				{
					// we got the data -> execute the command
					execute(s);
					s->state = ST_IDLE;
				}
				break;

			default:
				uartPrintStr_P(PSTR("err: unknown state\n"));
				// we are in an unknown state
				break;
		}
		
	}
	//uartPutc('\n');
}


void execute (struct sump_state* s)
{
	//uint8_t i;
	/* make the code more readable by renaming opcode and data */
	uint8_t opcode = s->opcode;
	uint32_t data32 = ((uint32_t)s->data[3])<<24 | \
					 ((uint32_t)s->data[2])<<16 | \
					 ((uint32_t)s->data[1])<<8 | ((uint32_t)s->data[0]);

	fprintf_P(&mystdout, PSTR("ex: 0x%02x\n"), opcode);
	
	if (opcode == OPCODE_RESET)
	{
		uartPrintStr_P(PSTR("cmd: reset\n"));
		waitForTrigger = 0;	
		hwDataReady = 0;
		LA_OPCODE_REG = opcode;
		return;
	}
	if (opcode == OPCODE_ID)
	{
		uartPrintStr_P(PSTR("cmd: id\n"));
		// if this is the connection from the active client we respond with a proper
		// id code. Otherwise send back a notification that the hardware is in use
		if (connectedClient == uip_conn)
		{
			if (send(s, ID_CODE_READY, 4) != 4)
			{
				uartPrintStr_P(PSTR("err: txBuffer overflow"));
				uip_close();
			}
			
		}
		else
		{
			if (send(s, ID_CODE_ERR0, 4) != 4)
			{
				uartPrintStr_P(PSTR("err: txBuffer overflow"));
				uip_close();
			}
		}
		return;
	}
	// other commands are only valid if they are sent from the connected client
	if (connectedClient != uip_conn)
		return;
	// if we are sampling right now other commands are not valid
	if (waitForTrigger)
		return;
	if (opcode == OPCODE_SIZE_BWD)
	{
		// client tells us how many bytes of pre-trigger data it wants
		preTrigLen = data32;
		uartPrintStr_P(PSTR("cmd: bwd\n"));
		//data32++;
		
	}
	if (opcode == OPCODE_SIZE_FWD)
	{
		// client tells us how many bytes of post-trigger data it wants
		postTrigLen = data32;
		uartPrintStr_P(PSTR("cmd: fwd\n"));
		data32--;		// we tell the hardware where to stop, so it takes fwd+1 samples
	}
	if (opcode == OPCODE_SET_READ_DEL)
	{
		/*
		// this is a legacy command, we could keep it for compatibilty
		postTrigLen = (((uint32_t)s->data[3])<<8) | ((uint32_t)s->data[2]);
		postTrigLen++;
		postTrigLen <<= 2;	// the value is in multiples of 4 samples
		preTrigLen  = (((uint32_t)s->data[1])<<8) | ((uint32_t)s->data[0]);
		preTrigLen++;
		preTrigLen <<= 2;
		// we now have the total number of samples in preTrigLen
		preTrigLen -= postTrigLen;
		// the hardware doesn't support this command any more -> set fwd and bwd
		LA_DATA_REG = postTrigLen;
		asm("nop");
		asm("nop");
		asm("nop");
		LA_OPCODE_REG = OPCODE_SIZE_FWD;
		// give the hardware some time to process the command
		for (i=0;i<15;i++)
			asm("nop");
		LA_DATA_REG = preTrigLen;
		asm("nop");
		asm("nop");
		asm("nop");
		LA_OPCODE_REG = OPCODE_SIZE_BWD;
		*/
		uartPrintStr_P(PSTR("The legacy cmd OPCODE_SET_READ_DEL (0x81) is not supported anymore.\n"));
		uip_close();
		return;
	}
	if (opcode == OPCODE_DATA_PORT)
	{
		// client tells where to send the sample data (in little endian)
		dataDestPort = ((uint16_t)s->data[1])<<8 | s->data[0];
		/*uartPrintStr("data port: ");
		uartPrintUInt16(dataDestPort);
		uartPutc('\n');*/
		fprintf_P(&mystdout, PSTR("data port: %u\n"),dataDestPort);
		return;
	}
	if (opcode == OPCODE_RESEND)
	{
		// the client wants a udp data packet resent because it missed it
		// this only makes sense if we have data in the sampling memory
		if (!hwDataReady)
			return;

		//fprintf_P(&mystdout, PSTR("cmd: resend, start: %08x") data32);
		
		// calc the start address of the packet in sampling memory
		uint32_t	startAdr;
		// Note: <<2 converts between words and bytes
		cli();	// note access to DRAM_XX has to be atomic
		startAdr =  DRAM_SAMPLING_PTR - (preTrigLen<<2) - (postTrigLen<<2);
		sei();
		// we now have the address of the first sample, add offset of requested sample
		startAdr += (data32<<2);
		cli();
		DRAM_SRC_ADDR = (startAdr&SAMPLE_MEM_ADDR_MASK) ;
		sei();
		// Note: the correct sample number has to apper in the packet. Therefore
		// we bias samplesSent and samplesToSend with requested index
		samplesSent = data32;	
		// calc the full length and subtract the offset
		samplesToSend = preTrigLen + postTrigLen - data32;
		// we never send more than 1 packet here -> limit the length
		if (samplesToSend > SAMPLES_PER_UDP_P)
			samplesToSend = SAMPLES_PER_UDP_P;
		samplesToSend += data32;	// add the offset, we now have the correct stop value
		/*uartPutc(' ');
		printHex32(samplesToSend);
		uartPutc('\n');*/

		// register a udp connection for sampling data transfer with the stack
		udpDataConn = uip_udp_new(&(uip_conn->ripaddr), htons(dataDestPort));
		if (udpDataConn != NULL)
			uip_udp_bind(udpDataConn, htons(dataDestPort));

		return;
	}
	if (opcode == OPCODE_ARM)
	{
		hwDataReady = 0;	// we start sampling -> the memory content is now invalid
		samplesSent = 0;	// until now we have not send any samples
		samplesToSend = preTrigLen + postTrigLen;	// total number of samples to send
		// register a udp connection for sample data transfer with the stack
		udpDataConn = uip_udp_new(&(uip_conn->ripaddr), htons(dataDestPort));
		if (udpDataConn != NULL)
			uip_udp_bind(udpDataConn, htons(dataDestPort));
		uartPrintStr("cmd: arm\n");
		uartPrintStr(" pre: ");
		printHex32(preTrigLen);
		uartPrintStr("  post: ");
		printHex32(postTrigLen);
		uartPutc('\n');
		waitForTrigger = 1;	// stop the timeout counter for this connection while we wait for the trigger
	}
	// pass the command to the hardware
	uartPrintStr_P(PSTR("hw ex:"));
	if (opcode & 0x80)
	{
		//fprintf_P(&mystdout, PSTR("%08x "), data32);
		cli();
		LA_DATA_REG = data32;
		sei();
	}
	//fprintf_P(&mystdout,PSTR("%02x\n"),opcode);
	cli();
	LA_OPCODE_REG = opcode;
	sei();
}


/* called by the stack when a new udp datagram arrived or polled as fast as possible
   for all registered udp connections */
void udpAppcall (void)
{
	uint32_t i;

	if (!configured)
	{
		// the dhcp client is active, we simply foreward all udp data to it
		handle_dhcp();
		return;
	}	

	if (uip_poll())
	{
		if ((uip_udp_conn == broadcastConn) && (timer_expired(&udpTimer)))
		{
			timer_reset(&udpTimer);
			if  (connectedClient==NULL)
			{
				// send an advertisement package 				
				struct advertisementPkg* p = (struct advertisementPkg*)uip_appdata;
				memcpy(p->name, "SumpLA", 6);
				memcpy(p->idCode, ID_CODE_READY, 4);
				p->port = HTONS(TCP_PORT);
				memcpy(&(p->deviceIpAddr), uip_hostaddr, 4);
				p->sampleMemLength = HTONL(SAMPLE_MEM_LENGTH);
				p->numChannels = HTONS(NUM_SAMPLING_CHANNELS);
				p->maxSamplingFrq = HTONS(MAX_SAMPLING_FRQ);	
				uip_udp_send(sizeof(struct advertisementPkg));
			}
		}
		else if ((uip_udp_conn == udpDataConn) && hwDataReady)
		{
			// Note: the dma module keeps track of the source address so we don't have to set it here
			//timer_reset(&udpTimer);
			if (samplesSent < samplesToSend)
			{
				uint16_t len;
				if ((samplesToSend-samplesSent)>=SAMPLES_PER_UDP_P)
					len = SAMPLES_PER_UDP_P;
				else
					len = samplesToSend-samplesSent;
				
				// copy the sample number of the first sample into the first word of the packet
				*((uint32_t*)uip_appdata) = samplesSent;
				// copy the number of samples into the packet
				((uint8_t*)uip_appdata)[4] = (uint8_t)(len & 0xff);
				((uint8_t*)uip_appdata)[5] = (uint8_t)((len>>8) & 0xff);
				
				//if (((uint16_t)(uip_appdata+6)) & 0x0003)
				//	uartPrintStr("ERR: appdata not on word boundary\n");

				/*printHex32(DRAM_SRC_ADDR);
				uartPutc(' ');
				printHex32(DRAM_SAMPLING_PTR);
				uartPutc(' ');
				printHex32((uint16_t)(uip_appdata)+6);
				uartPutc(' ');
				printHex32(DRAM_CFG_STAT);
				uartPutc(' ');*/

				// set the destination address
				// note: due to the WB-WB-bridge these accesses have to be atomic -> disable interrupts
				cli();
				DRAM_DEST_ADDR = (uint32_t)((uint16_t)uip_appdata+6);
				DRAM_TX_LEN = (uint32_t)len;
				sei();
				
				/*uartPutc('@');
				printHex32(DRAM_TX_LEN);
				uartPutc(' ');*/
				
				// wait until the dma transfer is done
				do
				{
					cli();
					i = DRAM_CFG_STAT;
					sei();	
				
					/*printHex32(DRAM_CFG_STAT);
					uartPutc(' ');
					printHex32(DRAM_TX_LEN);
					uartPutc(' ');*/
					// add some nops here so interrupts won't be dissable so much of the time
					asm("nop");
					asm("nop");
					asm("nop");
				}
				while (!((i) & (DRAM_DONE | DRAM_ERR)));
																
				cli();
				if (DRAM_CFG_STAT & DRAM_ERR)
					uartPutc('e');
				sei();

				// clear err and done flag
				cli();
				DRAM_CFG_STAT	= (uint32_t)(DRAM_DONE | DRAM_ERR);	
				sei();

				samplesSent += len;
				// tell the stack to send the packet
				uip_udp_send((len<<2)+6);
				//uartPutc('\n');
			}
			else
			{
				// we are done -> close udp socket
				uip_udp_remove(udpDataConn);
				/*uartPrintStr("cfg after last packet: ");
				printHex32(DRAM_CFG_STAT);
				uartPutc('\n');*/
			}
		}
	}
	if (uip_newdata())
	{
		uartPrintStr("rxed udp packet\n");
	}
}


/* sampling core interrupt */
ISR (INT3_vect)
{
	// the sampling memory (dram) is now valid
	hwDataReady = 1;
	waitForTrigger = 0;

	// set the dma source address to the beginning of the data section
	uint32_t	startAdr;
	startAdr =  DRAM_SAMPLING_PTR - (preTrigLen<<2) - (postTrigLen<<2);
	DRAM_SRC_ADDR = (startAdr&SAMPLE_MEM_ADDR_MASK) ;
}


/* initialize global variables and register connections with the stack */
void appInit (void)
{
	// register the TCP port we want to listen on
	uip_listen(HTONS(TCP_PORT));
	connectedClient = NULL;
	udpDataConn = NULL;
	hwDataReady = 0;
	waitForTrigger = 0;
	dataDestPort = DEFAULT_DATA_PORT;
	// register a udp connection to send our annotation broadcasts
	uip_ipaddr_t addr;
 	uip_ipaddr(&addr,255,255,255,255);
 	broadcastConn = uip_udp_new(&addr, HTONS(BROADCAST_PORT));
 	timer_set(&udpTimer, (uint32_t)CLOCK_SECOND);

	/*uartPrintStr("bc conn: ");
	printHex32(broadcastConn);
	uartPutc('\n');*/

	if(broadcastConn != 0) 
	{
		// ok the stack gave us a connection, set the destination port
   		uip_udp_bind(broadcastConn, HTONS(BROADCAST_PORT));
	}
	else
	{
		// this should never happen, it means the stack has not enough connections
		uartPrintStr("err: bc conn=0\n");
	}
	// ok, the application was configured
	configured = 1;	
}

/* reset the application, this might be necessary if the ethernet link is lost
   This function restores an unconfigured state similar to a hardware reset */
void appReset()
{
	/* reset all globals to their default reset value */
	cli();
	broadcastConn 	= 0;
	udpDataConn 	= 0;
	configured 		= 0;
	connectedClient = 0;
	preTrigLen 		= 0;	
	postTrigLen 	= 0;	
	dataDestPort 	= 0;		
	hwDataReady 	= 0;	
	waitForTrigger  = 0;
	samplesSent 	= 0;	
	samplesToSend 	= 0;			
	sei();

}

/* copy a message into the transmit buffer, transmission starts when the stack is ready to do so */
/* if the data is longer than the free buffer space only as many bytes as fit into the buffer will
   be copied. 
   returns: the number of bytes copied into the transmit buffer */
uint8_t send (struct sump_state* s, const void* data, uint8_t len)
{
	uint8_t i;
	int8_t nFree;
	// calculate the number of free bytes in the buffer
	nFree = s->txBufRd - s->txBufWr - 1;
	if (nFree < 0)
		nFree += BUFFER_LENGTH;

	if (len > nFree)
		len = nFree;

	for (i=0; i<len; i++)
	{
		s->txBuf[(s->txBufWr)++] = ((uint8_t*)data)[i];
		// check index overflow
		if (s->txBufWr >= BUFFER_LENGTH)
			s->txBufWr = 0;
	}
	return i;
}


/* end of file */
