/**************************************************************************************************************
--*
--*   B i t H o u n d   -   A n   F P G A   B a s e d   L o g i c   A n a l y z e r
--*
--*   Firmware 
--* 
--* Copyright 2010-2011   Mario Mauerer (MM), Lukas Schrittwieser (LS), ETH Zurich
--*
--*    This program is free software: you can redistribute it and/or modify
--*    it under the terms of the GNU General Public License as published by
--*    the Free Software Foundation, either version 3 of the License, or
--*    (at your option) any later version.
--*
--*    This program is distributed in the hope that it will be useful,
--*    but WITHOUT ANY WARRANTY; without even the implied warranty of
--*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--*    GNU General Public License for more details.
--*
--*    You should have received a copy of the GNU General Public License
--*    along with this program.  If not, see <http://www.gnu.org/licenses/>.
--*
--***************************************************************************************************************
--*
--* Change Log:
--*
--* Version 1.0 - 2011/08/17 - LS
--*   first release
--*
--*
--***************************************************************************************************************
--*
--* Ethernet driver implements. Handles the various data buffers and contorls the media access controller (MAC).
--*
--***************************************************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>

#include "eth.h"
#include "uart.h"
#include "io.h"

/* address of the PHY on the MDIO configuration interface */
#define PHY_MDIO_ADDR	0x07L

// memory address of mac configuration interface
#define ETH_BASE		0x8000

// macro to access an 32 bit config register (memory mapped)
#define REG32(x)		*((volatile uint32_t *)(x))

/* The packet buffers reside in the upper 12kB of data memory which is not seen by the compiler. 
   Therefore we define them ourselves using * const.
   The first NUM_ETH_RX_BUFFERS buffers are used for reception and transmission, the others are used for transmission only */
volatile uint8_t * const ethBuf[] = { 	(void*)0x1000, (void*)0x1600, (void*)0x1c00, (void*)0x2200, 
										(void*)0x2800, (void*)0x2e00, (void*)0x3400, (void*)0x3a00}; 
/* length of valid data in buffers */
volatile uint16_t 	ethBufLen[NUM_ETH_BUFFERS] = {0};

/* number (index) of buffer currently in transmission */
volatile int8_t 	transmittingBufNum = -1;

/* indicates that a buffer is to be sent */
volatile uint8_t 	bufFlags[NUM_ETH_BUFFERS] = {0};

/* link status, 1 means the phy has established a valid link */
volatile uint8_t	phyLinkState = 0;


void accessPhyReg(uint16_t * val, uint8_t addr, uint8_t write);



void ethInitPhy()
{
	cli();	// just to be sure

	uint16_t phyReg;
 	// tell the phy we don't support gigabit ethernet
	accessPhyReg(&phyReg, 9, 0);	// read register 9 (1000Base-T control reg)
	phyReg &= ~((1<<9) | (1<<8));	// clear bits 8 & 9 (i.e. do not advertise gigabit)
	accessPhyReg(&phyReg, 9,1);	// write register 9
	
	// software reset the phy and restart autonegotiation
	phyReg = (1<<15) | (1<<12);		// set reset and auto negotiate flag of ctrl reg
	accessPhyReg(&phyReg,0x00, 1);	// write config register
	do
	{
		accessPhyReg(&phyReg,0x00,0);	// read ctrl reg
	} while (phyReg & (1<<15));		// wait until reset is done


	// setup interrupt pin polirity
	/*accessPhyReg(&phyReg, 27, 0);   // read extended phy specific status reg
	phyReg |= (1<<10);	            // sel active high int polarity
	accessPhyReg(&phyReg, 27, 1);   */
	
	// enable link state interrupt
	accessPhyReg(&phyReg, 18, 0);
	phyReg |= ((uint16_t)1<<10);
	accessPhyReg(&phyReg, 18, 1);
}

void ethInit()
{
	uint8_t i;
	uint32_t *p;

	// note: due to the wishbone bridge accesses to the MAC have to be atomic
	// however this functions is called before interupts are enabled	
	cli();	// just to be sure..	

	//erase all interrupts
	REG32(ETH_BASE + ETH_INT_SOURCE) = ETH_RXC | ETH_TXC | ETH_BUSY | ETH_RXE | ETH_RXB | ETH_TXE | ETH_TXB; 
	
	//enable all interrupts
	REG32(ETH_BASE + ETH_INT_MASK) = ETH_RXC | ETH_TXC | ETH_BUSY | ETH_RXE | ETH_RXB | ETH_TXE | ETH_TXB;

	//set MAC ADDRESS //low word = mac address high word
	REG32(ETH_BASE + ETH_MAC_ADDR0) = ((uint32_t)(OWN_MAC_ADDRESS_2) << 24) | \
									  ((uint32_t)(OWN_MAC_ADDRESS_3) << 16) | \
									  ((uint32_t)(OWN_MAC_ADDRESS_4) << 8)  | \
									  ((uint32_t)(OWN_MAC_ADDRESS_5) << 0);
	//mac address rest
	REG32(ETH_BASE + ETH_MAC_ADDR1) = ((uint32_t)(OWN_MAC_ADDRESS_0) << 8) | \
									  ((uint32_t)(OWN_MAC_ADDRESS_1) << 0);	

	//configure TXBD0
	REG32(ETH_BASE + ETH_TXBD0H) = (uint32_t)((uint16_t)ethBuf[0]);		//address used for tx_data
	REG32(ETH_BASE + ETH_TXBD0L) = TX_READY;	//length = 0 | PAD & CRC = 1 | IRQ & WR = 1
	
	//configure RX buffer descriptors
	p = (uint32_t*)(ETH_BASE + ETH_RXBD0L);
	for (i=0; i<NUM_ETH_RX_BUFFERS; i++)
	{
		REG32(p+1) 	= (uint32_t)((uint16_t)ethBuf[i]);	// high word of buffer descriptor is the destination pointer
		REG32(p) 	= RX_READY;				// low word of buffer descriptor are flages: len = 0 | IRQ & WR = 1 | EMPTY = 1
		p += 2;	// a buffer descriptor is 2 words (=8 bytes) long		
	}
	// set wrap bit on last buffer descriptor
	REG32(p-2) = RX_READY | ETH_RXBD_WRAP;

	// enable transmitter and receiver, add CRC and Padding to frames, enable full duplex
	REG32(ETH_BASE + ETH_MODER) = ETH_TXEN | ETH_RXEN | ETH_PAD | ETH_CRCEN | ETH_FULLD;
	
	// enable reception of flow control frames
	REG32(ETH_BASE + ETH_CTRLMODER) = ETH_RXFLOW;
}


/* read or write a phy configuration register over the management intrface of the MAC 
	val: 	pointer to a uint16 with the data to write or where to stored the received value
	addr:	address of the phy register to be accesses
	write: 	a write access is performed if this is non zero
	*/
void accessPhyReg(uint16_t * val, uint8_t addr, uint8_t write)
{
	uint32_t a = (((uint16_t)addr)<<8) | PHY_MDIO_ADDR;

	// tell the mac the address of the register and the phy
	REG32(ETH_BASE + ETH_MIIADDRESS) 	= a;
	if (write)
	{
		REG32(ETH_BASE + ETH_MIITX_DATA)	= (uint32_t)(*val);
		// tell the mac to write to the phy
		REG32(ETH_BASE + ETH_MIICOMMAND)	= ETH_WCTRLDATA;
	}
	else
	{
		// tell the mac to read the phy register
		REG32(ETH_BASE + ETH_MIICOMMAND) 	= ETH_RSTAT;
	}
	/* wait until the command was processed */
	while (REG32(ETH_BASE + ETH_MIISTATUS)&ETH_MII_BUSY)
		asm("nop");
	
	if (!write)
	{
		a = REG32(ETH_BASE + ETH_MIIRX_DATA);
		(*val) = a & 0xFFFF;
	}
	//uartPrintStr("accessed MII reg\n");
}


/* starts the transmission of a frame of length bytes stored in buffer number bufferNum */
void ethSend(uint8_t bufferNum, uint16_t length)
{
	volatile uint8_t * b = ethBuf[bufferNum];
	
	// check teh bufferNumber
	if (bufferNum >= NUM_ETH_BUFFERS)
	{
		uartPrintStr("eth.c: invalid buffer num\n");
		return;
	}

	/*uartPrintStr("tx ");
	printHex(bufferNum);
	uartPutc(' ');
	printHex(((uint16_t)b)>>8);
	printHex(((uint16_t)b)&0xff);
	uartPutc(' ');
	printHex(((uint16_t)length)>>8);
	printHex(((uint16_t)length)&0xff);
	uartPutc(' ');
	printMac(b);
	uartPutc(' ');
	printMac(b+6);
	uartPutc('\n');*/
	
	// mark the buffer as ready for transmission
	cli();
	bufFlags[bufferNum] = BUF_TX;
	ethBufLen[bufferNum] = length;
	sei();

	// if we have a free buffer we start transmission 
	if (transmittingBufNum < 0)
	{
		cli();
		transmittingBufNum = bufferNum;
		REG32(ETH_BASE + ETH_TXBD0H) = (uint16_t)b;
		REG32(ETH_BASE + ETH_TXBD0L) = (( 0x0000FFFFL & ( ((uint32_t)length) ) ) << 16) | BD_SND;
		sei();
	}
}


/* returns the index of a free transmit buffer or -1 if no free buffer is available */
int8_t ethGetTxBuffer()
{
	/* we don't use the rx only buffers here, only the stack uses them to send data */
	int8_t ind=NUM_ETH_RX_BUFFERS;
	for (;ind<NUM_ETH_BUFFERS;ind++)
	{
		/* if the TX flag is zero this buffer is unused -> return its number */
		if (!(bufFlags[ind]&BUF_TX))
			return ind;
	}

	/* no buffer found */
	return -1;
}


/* returns the index of the first buffer containing a received data frame or -1 if no frame is available */
int8_t ethGetRxPacket()
{
	static int8_t ind=0;	// we check the buffers in a round robin way
	uint8_t i;
	/* Note: reading the 8 bit flag register doesn't need mutual exclusion as this is an atomic operation */
	for (i=0; i<NUM_ETH_RX_BUFFERS; i++)
	{
		ind++;
		if (ind>=NUM_ETH_RX_BUFFERS)
			ind = 0;
		if (bufFlags[ind]&BUF_RX)
		{
			//printHex(ind);
			return ind;
		}
	}
	return -1;
}


/* blocks until the specified buffer was sent, if ind=-1 it waits until all buffers were sent */
void ethWaitUntilSent (int8_t ind)
{
	int8_t i;
	if (ind >= 0)
	{
		/* reading only one byte of information doesn't require mutual exclusion because it is an atomic operation */
		while (bufFlags[ind]&BUF_TX)
			asm("nop");
	}
	else
	{
		for (i=0; i<NUM_ETH_BUFFERS; i++)
			ethWaitUntilSent(i);
	}
}


/* clears the RX flag from the specified buffer, this means the received data packet was processed */
void ethRxConsume (uint8_t ind)
{
	/* only the first few buffers are receive buffers, so only this can be ready for reception */
	if (ind >= NUM_ETH_RX_BUFFERS)	
		return;
	/* clear the receive flag (we don't touch the tx flag as the buffer might contain tx data now) */
	cli();
	uint8_t f = bufFlags[ind];
	f &= ~BUF_RX;
	/* if the buffer doesn't contain tx data it is ready for reception again */
	if (!(f&BUF_TX))
		REG32(ETH_BASE + ETH_RXBD0L + (ind*ETH_BD_LENGTH)) |= ETH_RXBD_EMPTY;
	bufFlags[ind] = f;	
	sei();
}


/* ask the phy about it's link state 
   return: 1 = link ok
           0 = link is down */
uint8_t ethGetLinkState()
{
	
	return phyLinkState;
	/*uint16_t i=0;
	// the link state changed, update the global variable
	accessPhyReg(&i, 17, 0); // 17 is the Phy specific status register 
	return (uint8_t)((i>>10)&0x01);*/
}

/* PHY interrupt, executed if the phy detected a link problem */
ISR(INT5_vect)
{
	uint16_t phyReg,i = 0;

	// read the interrupt status register
	accessPhyReg(&phyReg, 19, 0);

	if (phyReg&(1<<10))
	{
		// the link state changed, update the global variable
		accessPhyReg(&i, 17, 0); // 17 is the Phy specific status register 
		phyLinkState = (uint8_t)((i>>10)&0x01);
	}
	
	
}


/* MAC interrupt, called by hardware */
ISR(INT2_vect)
{
	// Note: we have to access the mac as 32bit device even if we only need 8 bits. So we must always use
	// the REG32 macro
    uint32_t s = REG32(ETH_BASE + ETH_INT_SOURCE);
    uint8_t source = (uint8_t)s;
	uint8_t i;    
	uint8_t t;

    if (source & ETH_TXB)
    {
		t = transmittingBufNum;	/* save time, transmittingBufNum is volatile */
		/* clear the TX flag of the buffer we sent */
		bufFlags[t] = 0;
		/* if this is a receive buffer it is now ready for new packets */
		if (t<NUM_ETH_RX_BUFFERS)	
			REG32(ETH_BASE + ETH_RXBD0L + (t*ETH_BD_LENGTH)) |= ETH_RXBD_EMPTY;
		/* assume we have no other packet to send */	
		transmittingBufNum=-1;	
		/* if another buffer is ready for transmission we start sending it */
		t++;	/* we try the buffer one after the last one first (round robin) */
		for (i=0; i<NUM_ETH_BUFFERS-1; i++)	/* we don't have to try all buffers because we know that the last one (the one we sent) is empty */
		{
			if (bufFlags[t]&BUF_TX)
			{
				REG32(ETH_BASE + ETH_TXBD0H) = (uint16_t)ethBuf[t];
				REG32(ETH_BASE + ETH_TXBD0L) = (( 0x0000FFFFL & ( ((uint32_t)ethBufLen[t]) ) ) << 16) | BD_SND;
				transmittingBufNum = t;
				break;
			}
			t++;	/* try next buffer */
			if (t>=NUM_ETH_BUFFERS)
				t=0;
		}
	}

    if (source & ETH_RXB)
    {
		/* check all reception buffers */
		for (i=0; i<NUM_ETH_RX_BUFFERS; i++)
		{
        	uint32_t bd = REG32(ETH_BASE + ETH_RXBD0L + (i*ETH_BD_LENGTH));
        	// check the empty and the seven error flags of the buffer descriptor
        	if ((bd & ETH_RXBD_EMPTY) == 0)
			{
				if ((bd & 0x7F) == 0)
				{
        			ethBufLen[i] = (REG32(ETH_BASE + ETH_RXBD0L + (i*ETH_BD_LENGTH)) >> 16); // buffer contains data, inform stack
					bufFlags[i] |= BUF_RX;
				}
				else
				{
					/* an error occured during reception, discard the frame */
					if (i<(NUM_ETH_RX_BUFFERS-1))
						REG32(ETH_BASE + ETH_RXBD0L + (i*ETH_BD_LENGTH)) = RX_READY;
					else
						/* note: the last buffer descriptor has the wrap bit set */
						REG32(ETH_BASE + ETH_RXBD0L + (i*ETH_BD_LENGTH)) = RX_READY | ETH_RXBD_WRAP;
				}
			}
		}       
    }
    // clear all interrupts
    REG32(ETH_BASE + ETH_INT_SOURCE) = ETH_RXC | ETH_TXC | ETH_BUSY | ETH_RXE | ETH_RXB | ETH_TXE | ETH_TXB;
}


